/*  einsc -- a compiler for eins
 *    Representation of the abstract syntax
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Printer.java,v 1.1 2005/03/29 14:01:35 dragos Exp $
 */

package einsc;

import einsc.Tree.*;


/** This class implements a pretty printer for abstract syntax
 *  trees in eins.
 */
public class Printer implements Visitor {

    /** Indentation string.
     */
    String step = "  ";

    /** indentation level.
     */
    int level = 0;

    /** Print a tree node.
     */
    public Printer print(Tree tree) {
        print("<" + tree.regNeed + ">");
        tree.apply(this);
        return this;
    }

    /** Print abstract syntax trees separated by commas.
     */
    Printer print(Tree[] trees) {
        for(int i = 0; i < trees.length; i++) {
            if (i > 0) print(", ");
            print(trees[i]);
        }
        return this;
    }

    /** Print a number.
     */
    Printer print(int value) {
        return print(String.valueOf(value));
    }

    /** Print a string.
     */
    Printer print(String value) {
        System.out.print(value);
        return this;
    }

    /** Print an end-of-line character and indent the following
     *  line appropriately.
     */
    Printer println() {
        System.out.println();
        for (int i = 0; i < level; i++)
            System.out.print(step);
        return this;
    }
    
    /** Increment the indentation level.
     */
    Printer indent() {
        level++;
        return this;
    }
    
    /** Decrement the indentation level.
     */
    Printer undent() {
        level--;
        return this;
    }

    //// VISITOR IMPLEMENTATION ////

    public void caseBad(Bad tree) {
        print("<bad>");
    }

    public void caseProgram(Program tree) {
        for (int i = 0; i < tree.decls.length; i++)
            print(tree.decls[i]).println();
        print(tree.main).println();
    }
    
    public void caseClassDef(ClassDef tree) {
        print("class ").print(tree.name);
        if (tree.fields.length == 0)
            print(" {}");
        else {
            print(" {").indent().println();
            for (int i = 0; i < tree.fields.length; i++) {
                print("val ").print(tree.fields[i]).print(";");
                if (i == tree.fields.length - 1)
                    undent();
                println();
            }
            print("}");
        }
    }

    public void caseFunDef(FunDef tree) {
        print("def ").print(tree.name).print("(").print(tree.params).print(")");
        print(": ").print(tree.restype).print(" = ").indent().println();
        print(tree.body).print(";").undent();
    }

    public void caseFormal(Formal tree) {
        print(tree.name).print(": ").print(tree.vartype);
    }
    
    public void caseIntType(IntType tree) {
        print("Int");
    }

    public void caseTypeIdent(TypeIdent tree) {
        print(tree.name);
    }

    public void caseFunType(FunType tree) {
        print("(").print(tree.paramtypes).print(") => ").print(tree.restype);
    }
    
    public void caseWhile(While tree) {
        print("while (").print(tree.cond).print(") {").indent().println();
        for (int i = 0; i < tree.stats.length; i++) {
            print(tree.stats[i]);
            if (i == tree.stats.length - 1)
                undent();
            println();
        }
        print("}");
    }

    public void caseVarDef(VarDef tree) {
        print("var ").print(tree.variable).print(" = ").indent();
        print(tree.init).print(";").undent();
    }

    public void caseAssign(Assign tree) {
        print("let ").print(tree.name + " = ").print(tree.expr).print(";");
    }

    public void caseDo(Do tree) {
        print("do ").print(tree.expr).print(";");
    }

    public void caseIf(If tree) {
        print("if (").indent().print(tree.cond).undent().print(")");
        indent().println().print(tree.thenp);
        undent().println().print("else");
        indent().println().print(tree.elsep).undent();
    }


    public void caseBinop(Binop tree) {
        print("(").print(tree.left);
        switch (tree.op) {
        case Tree.ADD:
            print(" + "); break;
        case Tree.SUB:
            print(" - "); break;
        case Tree.MUL:
            print(" * "); break;
        case Tree.DIV:
            print(" / "); break;
        case Tree.MOD:
            print(" % "); break;
        case Tree.EQ:
            print(" == "); break;
        case Tree.NE:
            print(" != "); break;
        case Tree.LT:
            print(" < "); break;
        case Tree.LE:
            print(" <= "); break;
        case Tree.GT:
            print(" > "); break;
        case Tree.GE:
            print(" >= "); break;
        case Tree.OR:
            print(" | "); break;
        default:
            throw new Error();
        }
        print(tree.right).print(")");
    }
    
    public void caseUnop(Unop tree) {
        switch (tree.op) {
        case Tree.NEG:
            print("-"); break;
        case Tree.NOT:
            print("!"); break;
        default:
            throw new Error();
        }
        print(tree.expr);
    }

    public void caseIdent(Ident tree) {
        print(tree.name);
    }
        
    public void caseNullLit(NullLit tree) {
        print("null");
    }
    
    public void caseIntLit(IntLit tree) {
        print(tree.value);
    }
    
    public void caseApply(Apply tree) {
        print(tree.fun).print("(").print(tree.args).print(")");
    }

    public void caseBlock(Block tree) {
        print("{").indent().println();
        for (int i = 0; i < tree.stats.length; i++)
            print(tree.stats[i]).println();
        print(tree.main).undent().println().print("}");
    }

    public void caseNew(New tree) {
        print("new ").print(tree.name).print("(").print(tree.args).print(")");
    }

    public void caseSelect(Select tree) {
        print(tree.prefix).print(".").print(tree.selector);
    }

}
