/*  einsc -- a compiler for eins
 *    Type representation
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Type.java,v 1.1 2005/03/29 14:01:32 dragos Exp $
 */

package einsc;


/** This class provides a representation for types in eins
 */
class Type {

    /** This singleton type is used when there is no type associated
     *  with a tree node.
     */
    public static final Type NO_TYPE = new Type();

    /** This type represents an erroneous type (caused by a typing
     *  error).
     */
    public static final Type BAD_TYPE = new Type();

    /** The type of the null object reference.
     */
    public static final Type NULL_TYPE = new Type();

    /** Type Int.
     */
    public static final Type INT_TYPE = new Type();

    /** This method can be used to check if two types are equal. It
     * must be overriden for object types and function types. It will
     * not be used for class types.
     */
    public boolean isSameType(Type that) {
        return
            (this == BAD_TYPE) ||
            (that == BAD_TYPE) ||
            (this == that);
    }

    /** This method checks if the provided type is a super type
     *  of the receiver.
     */
    public boolean isSubType(Type that) {
        return
            (this == BAD_TYPE) ||
            (that == BAD_TYPE) ||
            ((this == NULL_TYPE) && (that instanceof ObjType)) ||
            isSameType(that);
    }

    /** Return the less specific type of this type and 'that'.
     */
    public Type max(Type that) {
        if (isSubType(that))
            return that;
        if (that.isSubType(this))
            return this;
        return BAD_TYPE;
    }

    /** Returns a textual representation of this type.
     */
    public String toString() {
        if (this == NO_TYPE)
            return "<none>";
        if (this == BAD_TYPE)
            return "<bad>";
        if (this == NULL_TYPE)
            return "Null";
        if (this == INT_TYPE)
            return "Int";
        throw new Error();
    }

    /** Class types.
     */
    public static class ClassType extends Type {
        public final Scope fields;
        
        public ClassType(Scope fields) {
            this.fields = fields;
        }

        public String toString() {
            return "<class type>";
        }
    }

    /** Object types.
     */
    public static class ObjType extends Type {
        public final Symbol clazz;
        
        public ObjType(Symbol clazz) {
            this.clazz = clazz;
        }

        public boolean isSameType(Type that) {
            return (that instanceof ObjType) &&
                (clazz == ((ObjType) that).clazz);
        } 

        public String toString() {
            return clazz.name;
        }
    }

    /** Representation for function types.
     */
    public static class FunType extends Type {
        public final Type[] argTypes;
        public final Type resType;

        public FunType(Type[] argTypes, Type resType) {
            this.argTypes = argTypes;
            this.resType = resType;
        }

        public boolean isSameType(Type that) {
            return super.isSameType(that) ||
                ((that instanceof FunType) &&
                 sameArgTypes((FunType) that) &&
                 resType.isSameType(((FunType) that).resType));
        }

        private boolean sameArgTypes(FunType funType) {
            if (argTypes.length != funType.argTypes.length)
                return false;
            for (int i = 0; (i < argTypes.length); i++)
                if (!argTypes[i].isSameType(funType.argTypes[i]))
                    return false;
            return true;
        }

        public String toString() {
            StringBuffer buffer = new StringBuffer();
            buffer.append("(");
            for (int i = 0; i < argTypes.length; ++i) {
                if (i > 0)
                    buffer.append(", ");
                buffer.append(argTypes[i]);
            }
            buffer.append(") => ");
            buffer.append(resType);
            return buffer.toString();
        }
    }
}
